@@ -21,6 +21,8 @@ import datetime
 import json
 import logging
 import os
+import socket
+from concurrent.futures import ThreadPoolExecutor
 
 from ansible import __version__ as ansible_version
 from ansible.parsing.ajson import AnsibleJSONEncoder
@@ -106,6 +108,23 @@ options:
     ini:
       - section: ara
         key: api_timeout
+  argument_labels:
+    description: |
+        A list of CLI arguments that, if set, will be automatically applied to playbooks as labels.
+        Note that CLI arguments are not always named the same as how they are represented by Ansible.
+        For example, --limit is "subset", --user is "remote_user" but --check is "check".
+    type: list
+    default:
+      - remote_user
+      - check
+      - tags
+      - skip_tags
+      - subset
+    env:
+      - name: ARA_ARGUMENT_LABELS
+    ini:
+      - section: ara
+        key: argument_labels
   default_labels:
     description: A list of default labels that will be applied to playbooks
     type: list
@@ -142,6 +161,18 @@ options:
     ini:
       - section: ara
         key: ignored_files
+  callback_threads:
+    description:
+      - The number of threads to use in API client thread pools
+      - When set to 0, no threading will be used (default) which is appropriate for usage with sqlite
+      - Using threads is recommended when the server is using MySQL or PostgreSQL
+    type: integer
+    default: 0
+    env:
+      - name: ARA_CALLBACK_THREADS
+    ini:
+      - section: ara
+        key: callback_threads
 """
 
 
@@ -157,22 +188,30 @@ class CallbackModule(CallbackBase):
     def __init__(self):
         super(CallbackModule, self).__init__()
         self.log = logging.getLogger("ara.plugins.callback.default")
+        # These are configured in self.set_options
         self.client = None
+        self.callback_threads = None
+
         self.ignored_facts = []
         self.ignored_arguments = []
         self.ignored_files = []
 
         self.result = None
+        self.result_started = {}
+        self.result_ended = {}
         self.task = None
         self.play = None
         self.playbook = None
         self.stats = None
         self.file_cache = {}
         self.host_cache = {}
+        self.task_cache = {}
+        self.delegation_cache = {}
 
     def set_options(self, task_keys=None, var_options=None, direct=None):
         super(CallbackModule, self).set_options(task_keys=task_keys, var_options=var_options, direct=direct)
 
+        self.argument_labels = self.get_option("argument_labels")
         self.default_labels = self.get_option("default_labels")
         self.ignored_facts = self.get_option("ignored_facts")
         self.ignored_arguments = self.get_option("ignored_arguments")
@@ -193,9 +232,29 @@ class CallbackModule(CallbackBase):
             verify=False if insecure else True,
         )
 
+        # TODO: Consider un-hardcoding this and plumbing pool_maxsize to requests.adapters.HTTPAdapter.
+        #       In the meantime default to 4 so we don't go above requests.adapters.DEFAULT_POOLSIZE.
+        #       Otherwise we can hit "urllib3.connectionpool: Connection pool is full"
+        self.callback_threads = self.get_option("callback_threads")
+        if self.callback_threads > 4:
+            self.callback_threads = 4
+
+    def _submit_thread(self, threadpool, func, *args, **kwargs):
+        # Manages whether or not the function should be threaded to keep things DRY
+        if self.callback_threads:
+            # Pick from one of two thread pools (global or task)
+            threads = getattr(self, threadpool + "_threads")
+            threads.submit(func, *args, **kwargs)
+        else:
+            func(*args, **kwargs)
+
     def v2_playbook_on_start(self, playbook):
         self.log.debug("v2_playbook_on_start")
 
+        if self.callback_threads:
+            self.global_threads = ThreadPoolExecutor(max_workers=self.callback_threads)
+            self.log.debug("Global thread pool initialized with %s thread(s)" % self.callback_threads)
+
         content = None
 
         if playbook._file_name == "__adhoc_playbook__":
@@ -212,6 +271,23 @@ class CallbackModule(CallbackBase):
                 self.log.debug("Ignoring argument: %s" % argument)
                 cli_options[argument] = "Not saved by ARA as configured by 'ignored_arguments'"
 
+        # Retrieve and format CLI options for argument labels
+        argument_labels = []
+        for label in self.argument_labels:
+            if label in cli_options:
+                # Some arguments are lists or tuples
+                if isinstance(cli_options[label], tuple) or isinstance(cli_options[label], list):
+                    # Only label these if they're not empty
+                    if cli_options[label]:
+                        argument_labels.append("%s:%s" % (label, ",".join(cli_options[label])))
+                # Some arguments are booleans
+                elif isinstance(cli_options[label], bool):
+                    argument_labels.append("%s:%s" % (label, cli_options[label]))
+                # The rest can be printed as-is if there is something set
+                elif cli_options[label]:
+                    argument_labels.append("%s:%s" % (label, cli_options[label]))
+        self.argument_labels = argument_labels
+
         # Create the playbook
         self.playbook = self.client.post(
             "/api/v1/playbooks",
@@ -219,11 +295,12 @@ class CallbackModule(CallbackBase):
             arguments=cli_options,
             status="running",
             path=path,
-            started=datetime.datetime.now().isoformat(),
+            controller=socket.getfqdn(),
+            started=datetime.datetime.now(datetime.timezone.utc).isoformat(),
         )
 
         # Record the playbook file
-        self._get_or_create_file(path, content)
+        self._submit_thread("global", self._get_or_create_file, path, content)
 
         return self.playbook
 
@@ -235,9 +312,9 @@ class CallbackModule(CallbackBase):
         # Load variables to verify if there is anything relevant for ara
         play_vars = play._variable_manager.get_vars(play=play)["vars"]
         if "ara_playbook_name" in play_vars:
-            self._set_playbook_name(name=play_vars["ara_playbook_name"])
+            self._submit_thread("global", self._set_playbook_name, play_vars["ara_playbook_name"])
 
-        labels = self.default_labels
+        labels = self.default_labels + self.argument_labels
         if "ara_playbook_labels" in play_vars:
             # ara_playbook_labels can be supplied as a list inside a playbook
             # but it might also be specified as a comma separated string when
@@ -249,28 +326,40 @@ class CallbackModule(CallbackBase):
             else:
                 raise TypeError("ara_playbook_labels must be a list or a comma-separated string")
         if labels:
-            self._set_playbook_labels(labels=labels)
+            self._submit_thread("global", self._set_playbook_labels, labels)
 
         # Record all the files involved in the play
         for path in play._loader._FILE_CACHE.keys():
-            self._get_or_create_file(path)
+            self._submit_thread("global", self._get_or_create_file, path)
 
+        # Note: ansible-runner suffixes play UUIDs when running in serial so 34cff6f4-9f8e-6137-3461-000000000005 can
+        # end up being 34cff6f4-9f8e-6137-3461-000000000005_2. Remove anything beyond standard 36 character UUIDs.
+        # https://github.com/ansible-community/ara/issues/211
         # Create the play
         self.play = self.client.post(
             "/api/v1/plays",
             name=play.name,
             status="running",
-            uuid=play._uuid,
+            uuid=play._uuid[:36],
             playbook=self.playbook["id"],
-            started=datetime.datetime.now().isoformat(),
+            started=datetime.datetime.now(datetime.timezone.utc).isoformat(),
         )
 
         return self.play
 
+    def v2_playbook_on_handler_task_start(self, task):
+        self.log.debug("v2_playbook_on_handler_task_start")
+        # TODO: Why doesn't `v2_playbook_on_handler_task_start` have is_conditional ?
+        self.v2_playbook_on_task_start(task, False, handler=True)
+
     def v2_playbook_on_task_start(self, task, is_conditional, handler=False):
         self.log.debug("v2_playbook_on_task_start")
         self._end_task()
 
+        if self.callback_threads:
+            self.task_threads = ThreadPoolExecutor(max_workers=self.callback_threads)
+            self.log.debug("Task thread pool initialized with %s thread(s)" % self.callback_threads)
+
         pathspec = task.get_path()
         if pathspec:
             path, lineno = pathspec.split(":", 1)
@@ -283,33 +372,38 @@ class CallbackModule(CallbackBase):
         # Get task file
         task_file = self._get_or_create_file(path)
 
-        self.task = self.client.post(
-            "/api/v1/tasks",
-            name=task.get_name(),
-            status="running",
-            action=task.action,
-            play=self.play["id"],
-            playbook=self.playbook["id"],
-            file=task_file["id"],
-            tags=task.tags,
-            lineno=lineno,
-            handler=handler,
-            started=datetime.datetime.now().isoformat(),
-        )
+        # Get task
+        self.task = self._get_or_create_task(task, task_file["id"], lineno, handler)
 
         return self.task
 
+    def v2_runner_on_start(self, host, task):
+        # v2_runner_on_start was added in 2.8 so this doesn't get run for Ansible 2.7 and below.
+        self.result_started[host.get_name()] = datetime.datetime.now(datetime.timezone.utc).isoformat()
+
     def v2_runner_on_ok(self, result, **kwargs):
-        self._load_result(result, "ok", **kwargs)
+        self._submit_thread("task", self._load_result, result, "ok", **kwargs)
 
     def v2_runner_on_unreachable(self, result, **kwargs):
-        self._load_result(result, "unreachable", **kwargs)
+        self._submit_thread("task", self._load_result, result, "unreachable", **kwargs)
 
     def v2_runner_on_failed(self, result, **kwargs):
-        self._load_result(result, "failed", **kwargs)
+        self._submit_thread("task", self._load_result, result, "failed", **kwargs)
 
     def v2_runner_on_skipped(self, result, **kwargs):
-        self._load_result(result, "skipped", **kwargs)
+        self._submit_thread("task", self._load_result, result, "skipped", **kwargs)
+
+    def v2_runner_item_on_ok(self, result):
+        self._update_delegation_cache(result)
+
+    def v2_runner_item_on_failed(self, result):
+        self._update_delegation_cache(result)
+
+    def v2_runner_item_on_skipped(self, result):
+        pass
+        # result._task.delegate_to can end up being a variable from this hook, don't save it.
+        # https://github.com/ansible/ansible/issues/75339
+        # self._update_delegation_cache(result)
 
     def v2_playbook_on_stats(self, stats):
         self.log.debug("v2_playbook_on_stats")
@@ -320,15 +414,28 @@ class CallbackModule(CallbackBase):
 
     def _end_task(self):
         if self.task is not None:
-            self.client.patch(
-                "/api/v1/tasks/%s" % self.task["id"], status="completed", ended=datetime.datetime.now().isoformat()
+            self._submit_thread(
+                "task",
+                self.client.patch,
+                "/api/v1/tasks/%s" % self.task["id"],
+                status="completed",
+                ended=datetime.datetime.now(datetime.timezone.utc).isoformat(),
             )
+            if self.callback_threads:
+                # Flush threads before moving on to next task to make sure all results are saved
+                self.log.debug("waiting for task threads...")
+                self.task_threads.shutdown(wait=True)
+                self.task_threads = None
             self.task = None
 
     def _end_play(self):
         if self.play is not None:
-            self.client.patch(
-                "/api/v1/plays/%s" % self.play["id"], status="completed", ended=datetime.datetime.now().isoformat()
+            self._submit_thread(
+                "global",
+                self.client.patch,
+                "/api/v1/plays/%s" % self.play["id"],
+                status="completed",
+                ended=datetime.datetime.now(datetime.timezone.utc).isoformat(),
             )
             self.play = None
 
@@ -339,15 +446,24 @@ class CallbackModule(CallbackBase):
         else:
             status = "completed"
 
-        self.playbook = self.client.patch(
-            "/api/v1/playbooks/%s" % self.playbook["id"], status=status, ended=datetime.datetime.now().isoformat()
+        self._submit_thread(
+            "global",
+            self.client.patch,
+            "/api/v1/playbooks/%s" % self.playbook["id"],
+            status=status,
+            ended=datetime.datetime.now(datetime.timezone.utc).isoformat(),
         )
 
+        if self.callback_threads:
+            self.log.debug("waiting for global threads...")
+            self.global_threads.shutdown(wait=True)
+
     def _set_playbook_name(self, name):
         if self.playbook["name"] != name:
             self.playbook = self.client.patch("/api/v1/playbooks/%s" % self.playbook["id"], name=name)
 
     def _set_playbook_labels(self, labels):
+        # Only update labels if our cache doesn't match
         current_labels = [label["name"] for label in self.playbook["labels"]]
         if sorted(current_labels) != sorted(labels):
             self.log.debug("Updating playbook labels to match: %s" % ",".join(labels))
@@ -382,14 +498,69 @@ class CallbackModule(CallbackBase):
             self.host_cache[host] = self.client.post("/api/v1/hosts", name=host, playbook=self.playbook["id"])
         return self.host_cache[host]
 
+    def _get_or_create_task(self, task, file_id=None, lineno=None, handler=None):
+        # Note: The get_or_create is handled through the serializer of the API server.
+        task_uuid = str(task._uuid)[:36]
+        if task_uuid not in self.task_cache:
+            if None in (file_id, lineno, handler):
+                raise ValueError("file_id, lineno, and handler are required to create a task")
+
+            self.log.debug("Task not in cache, getting or creating: %s" % task)
+            self.task_cache[task_uuid] = self.client.post(
+                "/api/v1/tasks",
+                name=task.get_name(),
+                status="running",
+                action=task.action,
+                play=self.play["id"],
+                playbook=self.playbook["id"],
+                file=file_id,
+                tags=task.tags,
+                lineno=lineno,
+                handler=handler,
+                started=datetime.datetime.now(datetime.timezone.utc).isoformat(),
+            )
+
+        return self.task_cache[task_uuid]
+
+    def _update_delegation_cache(self, result):
+        # If the task is a loop and delegate_to is a variable, result._task.delegate_to can return the variable
+        # instead of it's value when using the v2_runner_on_* hooks.
+        # We're caching the actual host names here from v2_runner_item_on_* hooks.
+        # https://github.com/ansible/ansible/issues/75339
+        if result._task.delegate_to:
+            task_uuid = str(result._task._uuid[:36])
+            if task_uuid not in self.delegation_cache:
+                self.delegation_cache[task_uuid] = []
+            self.delegation_cache[task_uuid].append(result._task.delegate_to)
+
     def _load_result(self, result, status, **kwargs):
         """
         This method is called when an individual task instance on a single
         host completes. It is responsible for logging a single result to the
         database.
         """
+        hostname = result._host.get_name()
+        self.result_ended[hostname] = datetime.datetime.now(datetime.timezone.utc).isoformat()
+
         # Retrieve the host so we can associate the result to the host id
-        host = self._get_or_create_host(result._host.get_name())
+        host = self._get_or_create_host(hostname)
+
+        # If the task was delegated to another host, retrieve that too.
+        # Since a single task can be delegated to multiple hosts (ex: looping on a host group and using delegate_to)
+        # this must be a list of hosts.
+        delegated_to = []
+        # The value of result._task.delegate_to doesn't get templated if the task was skipped
+        # https://github.com/ansible/ansible/issues/75339#issuecomment-888724838
+        if result._task.delegate_to and status != "skipped":
+            task_uuid = str(result._task._uuid[:36])
+            if task_uuid in self.delegation_cache:
+                for delegated in self.delegation_cache[task_uuid]:
+                    delegated_to.append(self._get_or_create_host(delegated))
+            else:
+                delegated_to.append(self._get_or_create_host(result._task.delegate_to))
+
+        # Retrieve the task so we can associate the result to the task id
+        task = self._get_or_create_task(result._task)
 
         results = strip_internal_keys(module_response_deepcopy(result._result))
 
@@ -413,19 +584,20 @@ class CallbackModule(CallbackBase):
         self.result = self.client.post(
             "/api/v1/results",
             playbook=self.playbook["id"],
-            task=self.task["id"],
+            task=task["id"],
             host=host["id"],
-            play=self.task["play"],
+            delegated_to=[h["id"] for h in delegated_to],
+            play=task["play"],
             content=results,
             status=status,
-            started=self.task["started"],
-            ended=datetime.datetime.now().isoformat(),
+            started=self.result_started[hostname] if hostname in self.result_started else task["started"],
+            ended=self.result_ended[hostname],
             changed=result._result.get("changed", False),
             # Note: ignore_errors might be None instead of a boolean
             ignore_errors=kwargs.get("ignore_errors", False) or False,
         )
 
-        if self.task["action"] in ["setup", "gather_facts"] and "ansible_facts" in results:
+        if task["action"] in ["setup", "gather_facts"] and "ansible_facts" in results:
             self.client.patch("/api/v1/hosts/%s" % host["id"], facts=results["ansible_facts"])
 
     def _load_stats(self, stats):
@@ -433,7 +605,10 @@ class CallbackModule(CallbackBase):
         for hostname in hosts:
             host = self._get_or_create_host(hostname)
             host_stats = stats.summarize(hostname)
-            self.client.patch(
+
+            self._submit_thread(
+                "global",
+                self.client.patch,
                 "/api/v1/hosts/%s" % host["id"],
                 changed=host_stats["changed"],
                 unreachable=host_stats["unreachable"],
